/* m_list.c -- LIST command;
 * Copyright (c) 2001-2002 ForestNet IRC Network;
 * This is a free software; see file `license' for details.
 * $Id: m_list.c,v 1.13 2002/11/18 12:44:08 hex Exp $
 */

#include <string.h>

#include "ircd.h"
#include "channel.h"
#include "hash.h" /* For CHANNELHASHSIZE */
#include "send.h"

/* parv[0] = sender prefix, parv[1,2,3...] = Channels or list options. */

extern void channel_modes(aClient *cptr, char *mbuf, char *pbuf, aChannel *chptr);

/*
 * send_list
 *
 * The function which sends
 * The function which sends the actual /list output back to the user.
 * Operates by stepping through the hashtable, sending the entries back if
 * they match the criteria.
 * cptr = Local client to send the output back to.
 * numsend = Number (roughly) of lines to send back. Once this number has
 * been exceeded, send_list will finish with the current hash bucket,
 * and record that number as the number to start next time send_list
 * is called for this user. So, this function will almost always send
 * back more lines than specified by numsend (though not by much,
 * assuming CHANNELHASHSIZE is was well picked). So be conservative
 * if altering numsend };> -Rak
 */
void send_list(aClient *cptr, int numsend)
{
	int    hashptr, done = 0;
	aChannel *chptr;
	char   *p;

#define l cptr->lopt /* lazy shortcut */

	for (hashptr = l->starthash; hashptr < CHANNELHASHSIZE; hashptr++) {
		for (chptr = hash_get_chan_bucket(hashptr); chptr; chptr = chptr->hnextch) {
			if (!ShowChannel(cptr, chptr))
				continue;

			if (!l->showall && ((chptr->users <= l->usermin) ||
			   ((l->usermax == -1)?0:(chptr->users >= l->usermax)) ||
			   ((chptr->creationtime||1) <= l->chantimemin) ||
			   (chptr->topic_time < l->topictimemin) ||
			   (chptr->creationtime >= l->chantimemax) ||
			   (chptr->topic_time > l->topictimemax)))
			   continue;
			/* For now, just extend to topics as well. Use patterns starting
			   with # to stick to searching channel names only. -Donwulff */
			if (l->nolist &&
				(find_str_match_link(&(l->nolist), chptr->chname) ||
				find_str_match_link(&(l->nolist), chptr->topic)))
				continue;
			if (l->yeslist &&
				(!find_str_match_link(&(l->yeslist), chptr->chname) &&
				!find_str_match_link(&(l->yeslist), chptr->topic)))
				continue;

			/* insert channel modes, if any */
			p = buf;
			if (chptr->mode.mode) {
				*p++ = '[';
				channel_modes(cptr, p, NULL, chptr);
				for (; *p; p++);
				*p++ = ']';
				if (ShowChannel(cptr, chptr))
					*p++ = ' ';
				*p = '\0';
			}
			*p = '\0';

			if (!SecretChannel(chptr) || IsMember(cptr, chptr)) {
				if (p != buf)
					*p++ = ' ';
				*p = '\0';
				strlcat(buf, chptr->topic, sizeof(buf));
			}

			if (ShowChannel(cptr, chptr))
				sendto_one(cptr, rpl_str(RPL_LIST), me.name, cptr->name, chptr->chname, chptr->users, buf);

			if (--numsend == 0) /* Send to the end of the list and return  */
			done = 1;
		}

		if (done && (++hashptr < CHANNELHASHSIZE)) {
			l->starthash = hashptr;
			return;
		}
	}

	sendto_one(cptr, rpl_str(RPL_LISTEND), me.name, cptr->name);
	free_str_list(l->yeslist);
	free_str_list(l->nolist);
	MyFree(l);
	l = NULL;

	/* List finished, penalize by 10 seconds -Donwulff */
	if (!IsPrivileged(cptr))
		cptr->since += 10;

	return;
}


int m_list(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	aChannel *chptr;
	char   *name, *p = NULL;
	LOpts  *lopt;
	short  int showall = 0;
	Link   *yeslist = NULL, *nolist = NULL, *listptr;
	short  usermin = 0, usermax = -1;
	time_t currenttime = time(NULL);
	time_t chantimemin = 0, topictimemin = 0;
	time_t chantimemax, topictimemax;

	static char *usage[] = {
		"   Usage: /quote LIST options",
		"",
		"If you don't include any options, the default is to send you the",
		"entire unfiltered list of channels. Below are the options you can",
		"use, and what channels LIST will return when you use them.",
		">number  List channels with more than <number> people.",
		"<number  List channels with less than <number> people.",
		"C>number List channels created between now and <number> minutes ago.",
		"C<number List channels created earlier than <number> minutes ago.",
		"T>number List channels whose topics are older than <number> minutes",
		"         (Ie, they have not changed in the last <number> minutes.",
		"T<number List channels whose topics are not older than <number> minutes.",
		"*mask*   List channels that match *mask*",
		"!*mask*  List channels that do not match *mask*",
		NULL
	};


	/* None of that unregistered LIST stuff.  -- Barubary */
	if (check_registered(sptr)) return 0;

	/* I'm making the assumption it won't take over a day to transmit
	 * the list... -Rak
	 */
	chantimemax = topictimemax = currenttime + 86400;

	if ((parc == 2) && (!strcasecmp(parv[1], "?"))) {
		char **ptr = usage;

		for (; *ptr; ptr++)
			sendto_one(sptr, rpl_str(RPL_LISTSYNTAX), me.name, cptr->name, *ptr);

		return 0;
	}

	/* A list is already in process, for now we just interrupt the
	 * current listing, perhaps later we can allow stacked ones...
	 *  -Donwulff (Not that it's hard or anything, but I don't see
	 *             much use for it, beyond flooding)
	 */
	if (cptr->lopt) {
		free_str_list(cptr->lopt->yeslist);
		free_str_list(cptr->lopt->nolist);
		MyFree(cptr->lopt);
		cptr->lopt=NULL;

		sendto_one(sptr, rpl_str(RPL_LISTEND), me.name, cptr->name);
		/* Interrupted list, penalize 10 seconds */
		if(!IsPrivileged(sptr))
			sptr->since+=10;

		return 0;
	}

	sendto_one(sptr, rpl_str(RPL_LISTSTART), me.name, cptr->name);

	/* LIST with no arguements */
	if (parc < 2 || BadPtr(parv[1])) {
		lopt = (LOpts *)MyMalloc(sizeof(LOpts));
			if (!lopt)
				return 0;

		/* Changed to default to ignoring channels with only
		 * 1 person on, to decrease floods... -Donwulff
		 */
		memset(lopt, 0, sizeof(LOpts)); /* To be sure! */
		lopt->next = 0;
		lopt->yeslist = 0;
		lopt->nolist = 0;
		lopt->usermin = 2; /* Default */
		lopt->usermax = -1;
		lopt->chantimemax = lopt->topictimemax = currenttime + 86400;
		cptr->lopt = lopt;
		if (IsSendable(cptr))
			send_list(cptr, 64);
		return 0;
	}


	/*
	 * General idea: We don't need parv[0], since we can get that
	 * information from cptr.name. So, let's parse each element of
	 * parv[], setting pointer parv to the element being parsed.
	 */
	while (--parc) {
	   parv += 1;
	   if (BadPtr(parv)) /* Sanity check! */
		   continue;

	   name = strtoken(&p, *parv, ",");

	   while (name) {
		 switch (*name) {
		   case '>':
			   showall = 1;
			   usermin = (short)strtol(++name, (char **) 0, 10);
			   break;

		   case '<':
			   showall = 1;
			   usermax = (short)strtol(++name, (char **) 0, 10);
			   break;

		   case 't':
		   case 'T':
			   showall = 1;
			   switch (*++name) {
				   case '>':
					   topictimemax = currenttime - 60 *
									  strtol(++name, (char **) 0, 10);
					   break;

				   case '<':
					   topictimemin = currenttime - 60 * strtol(++name, (char **) 0, 10);
					   break;

				   case '\0':
					   topictimemin = 1;
					   break;

				   default:
					   sendto_one(sptr, err_str(ERR_LISTSYNTAX), me.name, cptr->name);
					   free_str_list(yeslist);
					   free_str_list(nolist);
					   sendto_one(sptr, rpl_str(RPL_LISTEND), me.name, cptr->name);

					   return 0;
			   }
			   break;

			   case 'c':
			   case 'C':
				   showall = 1;
				   switch (*++name) {
					   case '>':
						   chantimemin = currenttime - 60 *
										 strtol(++name, (char **) 0, 10);
						   break;

					   case '<':
						   chantimemax = currenttime - 60 *
										 strtol(++name, (char **) 0, 10);
						   break;

					   default:
						   sendto_one(sptr, err_str(ERR_LISTSYNTAX),
									  me.name, cptr->name);
						   free_str_list(yeslist);
						   free_str_list(nolist);
						   sendto_one(sptr, rpl_str(RPL_LISTEND),
									  me.name, cptr->name);
						   return 0;
				   }
				   break;

			   default: /* A channel or channel mask */

				   /*
					* new syntax: !channelmask will tell ircd to ignore
					* any channels matching that mask, and then
					* channelmask will tell ircd to send us a list of
					* channels only masking channelmask. Note: Specifying
					* a channel without wildcards will return that
					* channel even if any of the !channelmask masks
					* matches it.
					*/

				   if (*name == '!') {
					   showall = 1;
					   listptr = make_link();
					   listptr->next = nolist;
					listptr->value.cp = strdup(name + 1);
					   nolist = listptr;
				   }
				   else if (strchr(name, '*') || strchr(name, '?')) {
					   showall = 1;
					   listptr = make_link();
					   listptr->next = yeslist;
					listptr->value.cp = strdup(name);
					   yeslist = listptr;
				   }
				   else {
					 chptr = find_channel(name, NullChn);
					 if (chptr && ShowChannel(sptr, chptr))
						 sendto_one(sptr, rpl_str(RPL_LIST),
									me.name, cptr->name,
									ShowChannel(sptr,chptr) ? name : "*",
									chptr->users,
									chptr->topic);
				   }
		 } /* switch (*name) */
	   name = strtoken(&p, NULL, ",");
	   } /* while(name) */
	} /* while(--parc) */

	if (!showall || (chantimemin > currenttime) || (topictimemin > currenttime)) {
	  free_str_list(yeslist);
	  free_str_list(nolist);
	  sendto_one(sptr, rpl_str(RPL_LISTEND), me.name, cptr->name);
	  return 0;
	}

	lopt = (LOpts *)MyMalloc(sizeof(LOpts));

	lopt->showall = 0;
	lopt->next = NULL;
	lopt->yeslist = yeslist;
	lopt->nolist = nolist;
	lopt->starthash = 0;
	lopt->usermin = usermin;
	lopt->usermax = usermax;
	lopt->currenttime = currenttime;
	lopt->chantimemin = chantimemin;
	lopt->chantimemax = chantimemax;
	lopt->topictimemin = topictimemin;
	lopt->topictimemax = topictimemax;

	cptr->lopt = lopt;
	send_list(cptr, 64);

	return 0;
}
